"""
Copyright 2021-2021 The jdh99 Authors. All rights reserved.
dcom操作
Authors: jdh99 <jdh821@163.com>
"""

import tziot.config as config
import tziot.fpipe as fpipe
import tziot.standardlayer as standardlayer

import dcompy as dcom
import lagan
import utzpy as utz

# dcom是否初始化
is_dcom_init = False


def init_dcom():
    """初始化dcom.全局只应该初始化一次"""
    global is_dcom_init

    if is_dcom_init:
        return

    param = dcom.LoadParam()
    param.block_retry_max_num = config.dcom_retry_num
    param.block_retry_interval = config.dcom_retry_interval
    param.is_allow_send = fpipe.pipe_is_allow_send
    param.send = _dcom_send
    dcom.load(param)

    standardlayer.register_rx_observer(_deal_rx)
    is_dcom_init = True


def _dcom_send(protocol: int, pipe: int, dst_ia: int, data: bytearray):
    flp_frame = utz.bytes_to_flp_frame(data, True, 0)

    header = utz.StandardHeader()
    header.version = utz.PROTOCOL_VERSION
    header.frame_index = utz.generate_frame_index()
    header.payload_len = len(flp_frame)
    header.next_head = utz.HEADER_FLP
    header.hops_limit = 0xff
    header.src_ia = config.local_ia
    header.dst_ia = dst_ia

    standardlayer.send(flp_frame, header, pipe)


def _deal_rx(data: bytearray, standard_header: utz.StandardHeader, pipe: int):
    """处理标准层回调函数"""
    if standard_header.dst_ia != config.local_ia or standard_header.next_head != utz.HEADER_FLP:
        return

    body = utz.flp_frame_to_bytes(data)
    if body is None:
        lagan.warn(config.TAG, "flp frame to bytes failed!src ia:0x%x", standard_header.src_ia)
        return
    dcom.receive(config.PROTOCOL_NUM, pipe, standard_header.src_ia, body)



"""
Copyright 2021-2021 The jdh99 Authors. All rights reserved.
管道操作
Authors: jdh99 <jdh821@163.com>
"""

import tziot.config as config
import tziot.param as param

import socket
import threading
import lagan
import dcompy as dcom
import utzpy as utz

from typing import Callable

PIPE_NET = 0xffff


class _Api:
    # 是否允许发送.函数原型:func() bool
    is_allow_send = None  # type: Callable[[], bool]
    # 发送.函数原型:func(pipe: int, data: bytearray)
    send = None  # type: Callable[[int, bytearray], None]


_pipes = dict()
_pipe_num = 0
_socket = None
_observers = list()


def pipe_bind_net(ia: int, pwd: str, ip: str, port: int) -> int:
    """ 绑定网络管道.绑定成功后返回管道号"""
    global _socket

    if _socket is not None:
        lagan.warn(config.TAG, "already bind pipe net")
        return PIPE_NET

    config.local_pwd = pwd
    _socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    _socket.bind((ip, port))
    _bind(PIPE_NET, ia, _socket_tx, _socket_is_allow_send)
    threading.Thread(target=_socket_rx).start()

    return PIPE_NET


def _socket_rx():
    global _socket
    while True:
        data, address = _socket.recvfrom(config.FRAME_MAX_LEN)
        if len(data) == 0:
            continue
        lagan.info(config.TAG, 'udp rx:%r len:%d', address, len(data))
        lagan.print_hex(config.TAG, lagan.LEVEL_DEBUG, bytearray(data))
        pipe_receive(dcom.addr_to_pipe(address[0], address[1]), data)


def _socket_tx(pipe: int, data: bytearray):
    ip, port = dcom.pipe_to_addr(pipe)
    _socket.sendto(data, (ip, port))
    lagan.info(config.TAG, "udp send:ip:%s port:%d len:%d", ip, port, len(data))
    lagan.print_hex(config.TAG, lagan.LEVEL_DEBUG, data)


def _socket_is_allow_send() -> bool:
    return True


def pipe_receive(pipe: int, data: bytearray):
    """管道接收.pipe是发送方的管道号.如果是用户自己绑定管道,则在管道中接收到数据需回调本函数"""
    _notify_observers(pipe, data)


def _notify_observers(pipe: int, data: bytearray):
    global _observers

    for v in _observers:
        v(pipe, data)


def pipe_bind(ia: int, send, is_allow_send) -> int:
    """
    绑定管道.绑定成功后返回管道号
    :param ia: 设备单播地址
    :param send: 发送函数.格式:func(dst_pipe: int, data: bytearray)
    :param is_allow_send: 是否允许发送函数.格式:func() -> bool
    :return: 管道号
    """
    pipe = _get_pipe_num()
    _bind(pipe, ia, send, is_allow_send)
    return pipe


def _get_pipe_num() -> int:
    global _pipe_num

    _pipe_num += 1
    return _pipe_num


def _bind(pipe: int, ia: int, send, is_allow_send):
    config.local_ia = ia

    api = _Api()
    api.send = send
    api.is_allow_send = is_allow_send

    _pipes[pipe] = api


def pipe_is_allow_send(pipe: int) -> bool:
    if pipe >= PIPE_NET:
        pipe = PIPE_NET

    if pipe not in _pipes:
        return False
    return _pipes[pipe].is_allow_send()


def pipe_send(pipe: int, data: bytearray):
    if pipe == 0:
        return

    if pipe >= PIPE_NET:
        if PIPE_NET not in _pipes:
            return
        v = _pipes[PIPE_NET]
    else:
        if pipe not in _pipes:
            return
        v = _pipes[PIPE_NET]

    if pipe == PIPE_NET:
        if param.parent.ia == utz.IA_INVALID or not param.parent.is_conn:
            return
        v.send(param.parent.pipe, data)
    else:
        v.send(pipe, data)


def register_rx_observer(callback):
    """
    注册接收观察者
    :param callback: 回调函数格式:func(pipe: int, data: bytearray)
    """
    _observers.append(callback)



"""
Copyright 2021-2021 The jdh99 Authors. All rights reserved.
标准层处理模块
Authors: jdh99 <jdh821@163.com>
"""

import tziot.config as config
import tziot.fpipe as fpipe

import lagan
import utzpy as utz

_observers = list()


def init():
    fpipe.register_rx_observer(_rx)


def _rx(pipe: int, data: bytearray):
    """标准层接收"""
    header = _get_standard_header(data)
    if header is None:
        return

    _notify_observers(data[utz.NLV1_HEAD_LEN:], header, pipe)


def _get_standard_header(data: bytearray):
    header, offset = utz.bytes_to_standard_header(data)  # type: utz.StandardHeader, int
    if header is None or offset == 0:
        lagan.debug(config.TAG, 'get standard header failed:bytes to standard header failed')
        return None

    if header.version != utz.PROTOCOL_VERSION:
        lagan.debug(config.TAG, "get standard header failed:protocol version is not match:%d", header.version)
        return None

    if header.payload_len + offset != len(data):
        lagan.debug(config.TAG, "get standard header failed:payload len is not match:%d", header.payload_len)
        return None
    return header


def _notify_observers(data: bytearray, standard_header: utz.StandardHeader, pipe: int):
    global _observers

    for v in _observers:
        v(data, standard_header, pipe)


def register_rx_observer(callback):
    """
    注册接收观察者
    :param callback: 回调函数格式:func(data: bytearray, standardHeader: utz.StandardHeader, pipe: int)
    """
    _observers.append(callback)


def send(data: bytearray, standard_header: utz.StandardHeader, pipe: int):
    """基于标准头部发送"""
    data_len = len(data)
    if data_len > config.FRAME_MAX_LEN:
        lagan.error(config.TAG, "standard layer send failed!data len is too long:%d src ia:0x%x dst ia:0x%x", data_len,
                    standard_header.src_ia, standard_header.dst_ia)
        return

    if standard_header.payload_len != data_len:
        standard_header.payload_len = data_len

    frame = utz.standard_header_to_bytes(standard_header)
    frame += data
    fpipe.pipe_send(pipe, frame)



"""
Copyright 2021-2021 The jdh99 Authors. All rights reserved.
CMP协议解析处理
Authors: jdh99 <jdh821@163.com>
"""

import tziot.standardlayer as standardlayer
import tziot.config as config

import knocky as knock
import utzpy as utz
import lagan


def init():
    standardlayer.register_rx_observer(_deal_rx)


def _deal_rx(data: bytearray, standard_header: utz.StandardHeader, pipe: int):
    if standard_header.dst_ia != config.local_ia or standard_header.next_head != utz.HEADER_CMP:
        return

    payload = utz.flp_frame_to_bytes(data)
    if payload is None:
        lagan.warn(config.TAG, "parse cmp failed.flp frame to bytes failed")
        return

    if len(payload) == 0:
        lagan.warn(config.TAG, "parse cmp failed.payload len is wrong:%d", len(payload))
        return

    knock.call(utz.HEADER_CMP, payload[0], payload[1:])



"""
Copyright 2021-2021 The jdh99 Authors. All rights reserved.
申请父路由
Authors: jdh99 <jdh821@163.com>
"""

import tziot.config as config
import tziot.param as param
import tziot.standardlayer as standardlayer
import tziot.fpipe as fpipe
import tziot.fdcom as fdcom

import dcompy as dcom
import lagan
import utzpy as utz
import knocky as knock
import socket
import threading
import time


def init():
    knock.register(utz.HEADER_CMP, utz.CMP_MSG_TYPE_ASSIGN_SLAVE_ROUTER, deal_assign_slave_router)
    threading.Thread(target=_apply_thread).start()


def deal_assign_slave_router(req: bytearray, *args) -> (bytearray, bool):
    """处理分配从机帧.返回值是应答数据和应答标志.应答标志为false表示不需要应答"""
    if len(req) == 0:
        lagan.warn(config.TAG, "deal apply failed.payload len is wrong:%d", len(req))
        return None, False

    j = 0
    if req[j] != 0:
        lagan.warn(config.TAG, "deal apply failed.error code:%d", req[j])
        return None, False
    j += 1

    if len(req) != 16:
        lagan.warn(config.TAG, "deal apply failed.payload len is wrong:%d", len(req))
        return None, False

    param.parent.ia = utz.bytes_to_ia(req[j:j + utz.IA_LEN])
    j += utz.IA_LEN

    ip = socket.inet_ntoa(req[j:j + 4])
    j += 4
    port = (req[j] << 8) + req[j + 1]
    j += 2
    param.parent.pipe = dcom.addr_to_pipe(ip, port)

    lagan.info(config.TAG, "apply success.parent ia:0x%x ip:%s port:%d cost:%d", param.parent.ia, ip, port, req[j])
    return None, False


def _apply_thread():
    while True:
        # 如果网络通道不开启则无需申请
        if not fpipe.pipe_is_allow_send(fpipe.PIPE_NET):
            time.sleep(1)
            continue

        if fdcom.is_dcom_init and param.parent.ia == utz.IA_INVALID:
            lagan.info(config.TAG, "send apply frame")
            _send_apply_frame()

        if fdcom.is_dcom_init:
            time.sleep(10)
        else:
            time.sleep(1)


def _send_apply_frame():
    security_header = utz.SimpleSecurityHeader()
    security_header.next_head = utz.HEADER_CMP
    security_header.pwd = config.local_pwd
    payload = utz.simple_security_header_to_bytes(security_header)

    body = bytearray()
    body.append(utz.CMP_MSG_TYPE_REQUEST_SLAVE_ROUTER)
    body += utz.ia_to_bytes(param.parent.ia)
    body = utz.bytes_to_flp_frame(body, True, 0)

    payload += body

    header = utz.StandardHeader()
    header.version = utz.PROTOCOL_VERSION
    header.frame_index = utz.generate_frame_index()
    header.payload_len = len(payload)
    header.next_head = utz.HEADER_SIMPLE_SECURITY
    header.hops_limit = 0xff
    header.src_ia = config.local_ia
    header.dst_ia = config.core_ia

    standardlayer.send(payload, header, config.core_pipe)



"""
Copyright 2021-2021 The jdh99 Authors. All rights reserved.
公共参数
Authors: jdh99 <jdh821@163.com>
"""


class ParentInfo:
    """父路由信息"""

    def __init__(self):
        self.ia = 0
        self.pipe = 0
        self.cost = 0
        self.is_conn = False
        self.timestamp = 0


parent = ParentInfo()



"""
Copyright 2021-2021 The jdh99 Authors. All rights reserved.
配置文件
Authors: jdh99 <jdh821@163.com>
"""

import dcompy as dcom

TAG = 'tziot'

# 最大帧字节数
FRAME_MAX_LEN = 4096

PROTOCOL_NUM = 0

# 连接间隔.单位:s
CONN_INTERVAL = 30
# 连接超时时间.单位:s
CONN_TIMEOUT_MAX = 120

# 本机单播地址
local_ia = 0
local_pwd = ''

# 核心网参数
core_ia = 0x2141000000000002
core_ip = "115.28.86.171"
core_port = 12914
core_pipe = 0

# dcom参数
# dcom重发次数
dcom_retry_num = 5

# dcom重发间隔.单位:ms
dcom_retry_interval = 500


def init():
    global core_pipe
    core_pipe = dcom.addr_to_pipe(core_ip, core_port)


def config_core_param(ia: int, ip: str, port: int):
    """配置核心网参数"""
    global core_ia, core_ip, core_port, core_pipe
    core_ia = ia
    core_ip = ip
    core_port = port
    core_pipe = dcom.addr_to_pipe(ip, port)


def config_dcom_param(retry_num: int, retry_interval: int):
    """
    配置dcom参数
    :param retry_num: 重发次数
    :param retry_interval: 重发间隔.单位:ms
    """
    global dcom_retry_num, dcom_retry_interval

    if retry_num > 0:
        dcom_retry_num = retry_num
    if retry_interval > 0:
        dcom_retry_interval = retry_num



"""
Copyright 2021-2021 The jdh99 Authors. All rights reserved.
天泽物联网sdk
Authors: jdh99 <jdh821@163.com>
"""

import tziot.config as config
import tziot.param as param
import tziot.apply as apply
import tziot.fpipe as fpipe
import tziot.conn as conn
import tziot.parsecmp as parsecmp
import tziot.standardlayer as standardlayer
import tziot.fdcom as fdcom

import dcompy as dcom
import utzpy as utz

_is_first_run = True


def call(pipe: int, dst_ia: int, rid: int, timeout: int, req: bytearray) -> (bytearray, int):
    """
    RPC同步调用
    :param pipe: 通信管道
    :param dst_ia: 目标ia地址
    :param rid: 服务号
    :param timeout: 超时时间,单位:ms.为0表示不需要应答
    :param req: 请求数据.无数据可填bytearray()或者None
    :return: 返回值是应答字节流和错误码.错误码非0表示调用失败
    """
    if param.parent.ia == utz.IA_INVALID or not param.parent.is_conn:
        return None, dcom.SYSTEM_ERROR_RX_TIMEOUT

    if pipe >= fpipe.PIPE_NET:
        pipe = param.parent.pipe
    return dcom.call(config.PROTOCOL_NUM, pipe, dst_ia, rid, timeout, req)


def register(rid: int, callback):
    """
    注册DCOM服务回调函数
    :param rid: 服务号
    :param callback: 回调函数.格式: func(req bytearray) (bytearray, int)
    :return: 返回值是应答和错误码.错误码为0表示回调成功,否则是错误码
    """
    dcom.register(config.PROTOCOL_NUM, rid, callback)


def bind_pipe_net(ia: int, pwd: str, ip: str, port: int) -> int:
    """ 绑定网络管道.绑定成功后返回管道号"""
    global _is_first_run

    if _is_first_run:
        _is_first_run = False
        _init_system()
    return fpipe.pipe_bind_net(ia, pwd, ip, port)


def _init_system():
    config.init()
    apply.init()
    conn.init()
    parsecmp.init()
    standardlayer.init()
    fdcom.init_dcom()


def bind_pipe(ia: int, send, is_allow_send) -> int:
    """
    绑定管道.绑定成功后返回管道号
    :param ia: 设备单播地址
    :param send: 发送函数.格式:func(dst_pipe: int, data: bytearray)
    :param is_allow_send: 是否允许发送函数.格式:func() -> bool
    :return: 管道号
    """
    global _is_first_run

    if _is_first_run:
        _is_first_run = False
        _init_system()
    return fpipe.pipe_bind(ia, send, is_allow_send)


def str_to_bytearray(s: str, encoding: str = 'ascii') -> bytearray:
    """字符串转换成字节流.encoding是编码方式,默认是ascii码.如果有汉字需要选择utf-8等编码"""
    return bytearray(s.encode(encoding))


def bytearray_to_str(data: bytearray, encoding: str = 'ascii') -> str:
    """字节流转换成字符串.encoding是解码方式,默认是ascii码.如果有汉字需要选择utf-8等编码"""
    return data.decode(encoding)



"""
Copyright 2021-2021 The jdh99 Authors. All rights reserved.
连接父路由
Authors: jdh99 <jdh821@163.com>
"""

import tziot.config as config
import tziot.param as param
import tziot.fpipe as fpipe
import tziot.standardlayer as standardlayer

import knocky as knock
import utzpy as utz
import lagan
import threading
import time

# 最大连接次数.超过连接次数这回清除父路由IA地址,重连父路由
_CONN_NUM_MAX = 3

_conn_num = 0


def init():
    knock.register(utz.HEADER_CMP, utz.CMP_MSG_TYPE_ACK_CONNECT_PARENT_ROUTER, deal_ack_connect_parent_router)
    threading.Thread(target=_conn_thread).start()
    threading.Thread(target=_conn_timeout).start()


def deal_ack_connect_parent_router(req: bytearray, *args) -> (bytearray, bool):
    global _conn_num

    """dealAckConnectParentRouter 处理应答连接帧.返回值是应答数据和应答标志.应答标志为false表示不需要应答"""
    if len(req) == 0:
        lagan.warn(config.TAG, "deal conn failed.payload len is wrong:%d", len(req))
        return None, False

    j = 0
    if req[j] != 0:
        lagan.warn(config.TAG, "deal conn failed.error code:%d", req[j])
        return None, False
    j += 1

    if len(req) != 2:
        lagan.warn(config.TAG, "deal conn failed.payload len is wrong:%d", len(req))
        return None, False

    _conn_num = 0
    param.parent.is_conn = True
    param.parent.cost = req[j]
    param.parent.timestamp = int(time.time())
    lagan.info(config.TAG, "conn success.parent ia:0x%x cost:%d", param.parent.ia, param.parent.cost)
    return None, False


def _conn_thread():
    global _conn_num

    while True:
        # 如果网络通道不开启则无需连接
        if not fpipe.pipe_is_allow_send(fpipe.PIPE_NET):
            time.sleep(1)
            continue

        if param.parent.ia != utz.IA_INVALID:
            _conn_num += 1
            if _conn_num > _CONN_NUM_MAX:
                _conn_num = 0
                param.parent.ia = utz.IA_INVALID
                lagan.warn(config.TAG, "conn num is too many!")
                continue
            lagan.info(config.TAG, "send conn frame")
            _send_conn_frame()

        if param.parent.ia == utz.IA_INVALID:
            time.sleep(1)
        else:
            time.sleep(config.CONN_INTERVAL)


def _send_conn_frame():
    security_header = utz.SimpleSecurityHeader()
    security_header.next_head = utz.HEADER_CMP
    security_header.pwd = config.local_pwd
    payload = utz.simple_security_header_to_bytes(security_header)

    body = bytearray()
    body.append(utz.CMP_MSG_TYPE_CONNECT_PARENT_ROUTER)
    # 前缀长度
    body.append(64)
    # 子膜从机固定单播地址
    body += bytearray(utz.IA_LEN)
    # 开销值
    body.append(0)
    body = utz.bytes_to_flp_frame(body, True, 0)

    payload += body

    header = utz.StandardHeader()
    header.version = utz.PROTOCOL_VERSION
    header.frame_index = utz.generate_frame_index()
    header.payload_len = len(payload)
    header.next_head = utz.HEADER_SIMPLE_SECURITY
    header.hops_limit = 0xff
    header.src_ia = config.local_ia
    header.dst_ia = config.core_ia

    standardlayer.send(payload, header, param.parent.pipe)


def _conn_timeout():
    while True:
        if param.parent.ia == utz.IA_INVALID or not param.parent.is_conn:
            time.sleep(1)
            continue

        if int(time.time()) - param.parent.timestamp > config.CONN_TIMEOUT_MAX:
            param.parent.ia = utz.IA_INVALID
            param.parent.is_conn = False

        time.sleep(1)


def is_conn() -> bool:
    """是否连接核心网"""
    return param.parent.ia != utz.IA_INVALID and param.parent.is_conn
